static const char rcsid[] =
"$Id: mus_cdb.c,v 1.7 2002/03/26 06:41:47 schwartz Exp $";

#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include "util.h"
#include "seq.h"

#include "cdb.h"
#include "mus_cdb.h"

/* DB_DIR contains a number of fasta files, and a cdb file.  The cdb
   file maps strings like "gnl|ti|5" to "offset file".
   
   (The offset is decimal; maybe hex would be better?)

   To fetch a mus read, we look up the first token of its fasta header,
   seq_open the designated fasta file, seek to the right spot, and
   seq_read the data.   To avoid lots of open/close syscalls, we keep
   an array of open SEQ's, and remember which files they are connected to.

   (Unix has an open file limit, on the order of 1000, but there are
   less than 100 mfa files right now, so we are fine.)
*/

#define DB_DIR "/usr/local/align/data/mus/clipped"

enum {
 TOKLEN = 4096,
 KEYLEN = TOKLEN+32,
 PATHLEN = 8192,
 NMICE = 512,
 VBUFSIZE = 128*1024,
};

/* XXX - not reentrant, not thread safe. */
// -- const once initialized
static char cdb_dir[PATHLEN];
static char fasta_dir[PATHLEN];
static struct cdb c;
// -- scratch
static char key[KEYLEN];
static char path[PATHLEN];
// -- open file cache
static SEQ *mice[NMICE];

static unsigned char *mus_get(char *, unsigned int *);

void big_vbuf(FILE *fp)
{
	fflush(fp);
	setvbuf(fp, ckalloc(VBUFSIZE), _IOFBF, VBUFSIZE);
}

void mus_cdb_init(const char *d)
{
    mus_cdb_init2(d,0);
}

void mus_cdb_init2(const char *d, const char *f)
{
	int fd, n;

	if (d == 0) d = DB_DIR;
	if (f == 0) f = d;

	n = snprintf(fasta_dir, PATHLEN, "%s", f);
	if (n >= PATHLEN)
	    fatalf("mus_cdb_init: dir too long: '%s'", f);

	n = snprintf(cdb_dir, PATHLEN, "%s", d);
	if (n >= PATHLEN)
	    fatalf("mus_cdb_init: dir too long: '%s'", d);

	n = snprintf(path, PATHLEN, "%s/%s", cdb_dir, "cdb");
	if (n >= PATHLEN)
	    fatalf("mus_cdb_init: path too long: '%s/%s'", cdb_dir, "cdb");

	fd = open(path, 0);
	if (fd < 0)
	    fatalfr("mus_cdb_init: open (%s)", path);
	
	cdb_init(&c,fd);
}

char *fasta_key(char *key, char *h)
{
	int n;
	/* copy first token of h to tok */
	if (*h == '>') ++h;
	n = strcspn(h, " ");
	if (n >= TOKLEN)
	    fatalf("token too long (%s)", h);
	snprintf(key, KEYLEN, "%.*s", n, h); /* XXX - check return */
	return key;
}

unsigned char *mus_cdb_seq(char *h) {
	return mus_cdb_seq2(h, 0);
}

unsigned char *mus_cdb_seq2(char *h, unsigned int *len)
{
	unsigned int dlen;
	unsigned long dpos; /* int32 */
	unsigned char buf[8192];
	unsigned char *addr = buf;

	if (!h) h = "";
	fasta_key(key, h);

	switch (cdb_find(&c,key,strlen(key))) {
	 case  0: fatalf("cannot find contig {%s}", h);
	 case -1: fatalfr("cdb_find {%s}", key);
	}
	dlen = cdb_datalen(&c);
	dpos = cdb_datapos(&c);
	
	// addr = ckalloc(dlen+1);
	if (dlen+1 > sizeof(buf))
	    fatalf("dlen=%d > sizeof(buf)=%d\n", dlen+1, sizeof(buf));
	if (cdb_read(&c,addr,dlen,dpos) < 0)
	    fatalfr("cdb read error");
	addr[dlen] = 0;

	return mus_get(addr,len);
}

static unsigned char *mus_get(char *loc, unsigned int *len)
{
	SEQ *s;
	unsigned long pos;
	char *p;
	int n;

	/* loc looks like "%d %s" */
	pos = strtoul(loc, &p, 10); /* XXX - check? */
	while (*p == ' ') ++p;
	if (*p == 0) fatalf("bad file in loc: {%s}", loc);
	n = snprintf(path, PATHLEN, "%s/%s", fasta_dir, p);
        if (n >= PATHLEN)
            fatalf("mus_get: path too long: '%s/%s'", fasta_dir, p);

	/* open path, or reuse old SEQ */
	for (n=0; n < NMICE; ++n) {
		if (mice[n] == 0) {
		    // fprintf(stderr, "path={%s}\n", path);
		    mice[n] = seq_open(path);
		    //big_vbuf(mice[n]->fp);
		    break;
		} else if (strcmp(SEQ_NAME(mice[n]), path) == 0)
		    break;
	}
	if (n == NMICE)
		fatalf("mouse table full (%d)", NMICE);
	s = mice[n];

	//fprintf(stderr, "n=%d pos=%ld path={%s}\n", n, pos, path);
	if (fseek(s->fp, pos, SEEK_SET))
	    fatalfr("fseek", path);
	if (!seq_read(s))
	    fatalf("seq_read(%s)", path);
	// fprintf(stderr, "hdr={%s}\n", SEQ_HEAD(s));
	p = SEQ_CHARS(s);

	//p = copy_string(p);  // XXX - do we want to do this, or let the caller?
	s->seq = 0; // tell seq_read not to free p.
	if (len) *len = SEQ_LEN(s);
	return p;
}

static int lookups = 0;
static int hits = 0;
void mus_cdb_terminate(void)
{
	cdb_free(&c);
	/* could close mice[] files, too */
}

// Yet another hash table
// This one is for storing smallish contigs that might
// get read a lot.  Usefulness uncertain.
enum { HASHMAX = 5000 };
struct keyval {
    struct keyval *next;
    char *key;
    char *val;
};
static struct keyval *cache[HASHMAX];

void mus_cache_stats(void)
{
	fprintf(stderr, "cache hits=%d/%d\n", hits, lookups);
}

static unsigned int hash(char *s)
{
	unsigned int n = 5381;
	while (*s) {
	    n += n << 5;
	    n ^= *s++;
	}
	return n%HASHMAX;
}

unsigned char *mus_cached_seq(char *s)
{
	char key[KEYLEN];
	unsigned int x;
	struct keyval *kv;
	int i=0;

	++lookups;
	fasta_key(key, s);
	x = hash(key);
	for (kv = cache[x]; kv; kv = kv->next) {
	    if (kv->key && !strcmp(key, kv->key))
		break; 
	    ++i;
	}
	if (kv) ++hits;
	else {
	    kv = ckalloc(sizeof(struct keyval));
	    kv->next = cache[x];
	    kv->key = copy_string(key);
	    kv->val = mus_cdb_seq(key);
	    cache[x] = kv;
	}
	return copy_string(kv->val); // XXX - here or there?
}

#ifdef TESTING
static int len(char *s)
{
	if (!s) return -1;
	return strlen(s);
}

static void kv(char *k)
{
	//char *v = mus_cdb_seq(k);
	char *v = mus_cached_seq(k);
	printf("key=%d{%s}\nval=%d{%s}\n", len(k), k, len(v), v);
	free(v);
}

int main()
{
	char buf[TOKLEN];

	mus_cdb_init();

	while (fgets(buf, TOKLEN, stdin)) {
	    if (buf[0]) buf[strlen(buf)-1]=0;
	    kv(buf);
	}
	// kv(">gnl|ti|5 foo bar");
	// kv(">gnl|ti|5555 zot ");
	// kv(">gnl|ti|123456");
	// kv(">gnl|ti|654321");

	mus_cdb_terminate();
	return 0;
}
#endif
